"""
This package is for the characterisation and reporting of geometric distortion in MRI
Please see here for detailed documentation
https://acrf-image-x-institute.github.io/MRI_DistortionQA/index.html
"""

__version__ = '0.10.2'
from .calculate_harmonics import calculate_harmonics
